from __future__ import annotations

import copy
import logging
import os
import string
import time
from collections.abc import Iterable, Sequence
from datetime import timedelta
from pathlib import Path
from typing import (
    Annotated,
    Any,
    ClassVar,
    Literal,
)

import nshconfig as C
import numpy as np
from lightning.fabric.plugins.precision.precision import _PRECISION_INPUT
from lightning.pytorch.accelerators import Accelerator
from lightning.pytorch.callbacks.callback import Callback
from lightning.pytorch.loggers import Logger
from lightning.pytorch.plugins import _PLUGIN_INPUT
from lightning.pytorch.profilers import Profiler
from lightning.pytorch.strategies.strategy import Strategy
from typing_extensions import TypeAliasType, TypedDict, override

from .._directory import DirectoryConfig
from .._hf_hub import HuggingFaceHubConfig
from ..callbacks import (
    BestCheckpointCallbackConfig,
    CallbackConfig,
    EarlyStoppingCallbackConfig,
    LastCheckpointCallbackConfig,
    NormLoggingCallbackConfig,
    OnExceptionCheckpointCallbackConfig,
)
from ..callbacks.base import CallbackConfigBase
from ..callbacks.debug_flag import DebugFlagCallbackConfig
from ..callbacks.log_epoch import LogEpochCallbackConfig
from ..callbacks.lr_monitor import LearningRateMonitorConfig
from ..callbacks.metric_validation import MetricValidationCallbackConfig
from ..callbacks.rlp_sanity_checks import RLPSanityChecksCallbackConfig
from ..callbacks.shared_parameters import SharedParametersCallbackConfig
from ..loggers import (
    CSVLoggerConfig,
    LoggerConfig,
    TensorboardLoggerConfig,
    WandbLoggerConfig,
)
from ..loggers.actsave import ActSaveLoggerConfig
from ..loggers.base import LoggerConfigBase
from ..metrics._config import MetricConfig
from ..profiler import ProfilerConfig
from ..util._environment_info import EnvironmentConfig
from .accelerator import AcceleratorConfig, AcceleratorLiteral
from .plugin import PluginConfig
from .strategy import StrategyConfig, StrategyLiteral

log = logging.getLogger(__name__)


class GradientClippingConfig(C.Config):
    enabled: bool = True
    """Enable gradient clipping."""
    value: int | float
    """Value to use for gradient clipping."""
    algorithm: Literal["value", "norm"] = "norm"
    """Norm type to use for gradient clipping."""


CheckpointCallbackConfig = TypeAliasType(
    "CheckpointCallbackConfig",
    Annotated[
        BestCheckpointCallbackConfig
        | LastCheckpointCallbackConfig
        | OnExceptionCheckpointCallbackConfig,
        C.Field(discriminator="name"),
    ],
)


class CheckpointSavingConfig(CallbackConfigBase):
    enabled: bool = True
    """Enable checkpoint saving."""

    checkpoint_callbacks: Sequence[CheckpointCallbackConfig] = [
        BestCheckpointCallbackConfig(throw_on_no_metric=False),
        LastCheckpointCallbackConfig(),
        OnExceptionCheckpointCallbackConfig(),
    ]
    """Checkpoint callback configurations."""

    def disable_(self):
        self.enabled = False
        return self

    def should_save_checkpoints(self, trainer_config: TrainerConfig):
        if not self.enabled:
            return False

        if trainer_config.fast_dev_run:
            return False

        return True

    @override
    def create_callbacks(self, trainer_config: TrainerConfig):
        if not self.should_save_checkpoints(trainer_config):
            return

        for callback_config in self.checkpoint_callbacks:
            yield from callback_config.create_callbacks(trainer_config)


class LightningTrainerKwargs(TypedDict, total=False):
    accelerator: str | Accelerator
    """Supports passing different accelerator types ("cpu", "gpu", "tpu", "ipu", "hpu", "mps", "auto")
    as well as custom accelerator instances."""

    strategy: str | Strategy
    """Supports different training strategies with aliases as well custom strategies.
    Default: ``"auto"``.
    """

    devices: list[int] | str | int
    """The devices to use. Can be set to a positive number (int or str), a sequence of device indices
    (list or str), the value ``-1`` to indicate all available devices should be used, or ``"auto"`` for
    automatic selection based on the chosen accelerator. Default: ``"auto"``.
    """

    num_nodes: int
    """Number of GPU nodes for distributed training.
    Default: ``1``.
    """

    precision: _PRECISION_INPUT | None
    """Double precision (64, '64' or '64-true'), full precision (32, '32' or '32-true'),
    16bit mixed precision (16, '16', '16-mixed') or bfloat16 mixed precision ('bf16', 'bf16-mixed').
    Can be used on CPU, GPU, TPUs, HPUs or IPUs.
    Default: ``'32-true'``.
    """

    logger: Logger | Iterable[Logger] | bool | None
    """Logger (or iterable collection of loggers) for experiment tracking. A ``True`` value uses
    the default ``TensorBoardLogger`` if it is installed, otherwise ``CSVLogger``.
    ``False`` will disable logging. If multiple loggers are provided, local files
    (checkpoints, profiler traces, etc.) are saved in the ``log_dir`` of the first logger.
    Default: ``True``.
    """

    callbacks: list[Callback] | Callback | None
    """Add a callback or list of callbacks.
    Default: ``None``.
    """

    fast_dev_run: int | bool
    """Runs n if set to ``n`` (int) else 1 if set to ``True`` batch(es)
    of train, val and test to find any bugs (ie: a sort of unit test).
    Default: ``False``.
    """

    max_epochs: int | None
    """Stop training once this number of epochs is reached. Disabled by default (None).
    If both max_epochs and max_steps are not specified, defaults to ``max_epochs = 1000``.
    To enable infinite training, set ``max_epochs = -1``.
    """

    min_epochs: int | None
    """Force training for at least these many epochs. Disabled by default (None).
    """

    max_steps: int
    """Stop training after this number of steps. Disabled by default (-1). If ``max_steps = -1``
    and ``max_epochs = None``, will default to ``max_epochs = 1000``. To enable infinite training, set
    ``max_epochs`` to ``-1``.
    """

    min_steps: int | None
    """Force training for at least these number of steps. Disabled by default (``None``).
    """

    max_time: str | timedelta | dict[str, int] | None
    """Stop training after this amount of time has passed. Disabled by default (``None``).
    The time duration can be specified in the format DD:HH:MM:SS (days, hours, minutes seconds), as a
    :class:`datetime.timedelta`, or a dictionary with keys that will be passed to
    :class:`datetime.timedelta`.
    """

    limit_train_batches: int | float | None
    """How much of training dataset to check (float = fraction, int = num_batches).
    Default: ``1.0``.
    """

    limit_val_batches: int | float | None
    """How much of validation dataset to check (float = fraction, int = num_batches).
    Default: ``1.0``.
    """

    limit_test_batches: int | float | None
    """How much of test dataset to check (float = fraction, int = num_batches).
    Default: ``1.0``.
    """

    limit_predict_batches: int | float | None
    """How much of prediction dataset to check (float = fraction, int = num_batches).
    Default: ``1.0``.
    """

    overfit_batches: int | float
    """Overfit a fraction of training/validation data (float) or a set number of batches (int).
    Default: ``0.0``.
    """

    val_check_interval: int | float | None
    """How often to check the validation set. Pass a ``float`` in the range [0.0, 1.0] to check
    after a fraction of the training epoch. Pass an ``int`` to check after a fixed number of training
    batches. An ``int`` value can only be higher than the number of training batches when
    ``check_val_every_n_epoch=None``, which validates after every ``N`` training batches
    across epochs or during iteration-based training.
    Default: ``1.0``.
    """

    check_val_every_n_epoch: int | None
    """Perform a validation loop every after every `N` training epochs. If ``None``,
    validation will be done solely based on the number of training batches, requiring ``val_check_interval``
    to be an integer value.
    Default: ``1``.
    """

    num_sanity_val_steps: int | None
    """Sanity check runs n validation batches before starting the training routine.
    Set it to `-1` to run all batches in all validation dataloaders.
    Default: ``2``.
    """

    log_every_n_steps: int | None
    """How often to log within steps.
    Default: ``50``.
    """

    enable_checkpointing: bool | None
    """If ``True``, enable checkpointing.
    It will configure a default ModelCheckpoint callback if there is no user-defined ModelCheckpoint in
    :paramref:`~lightning.pytorch.trainer.trainer.Trainer.callbacks`.
    Default: ``True``.
    """

    enable_progress_bar: bool | None
    """Whether to enable to progress bar by default.
    Default: ``True``.
    """

    enable_model_summary: bool | None
    """Whether to enable model summarization by default.
    Default: ``True``.
    """

    accumulate_grad_batches: int
    """Accumulates gradients over k batches before stepping the optimizer.
    Default: 1.
    """

    gradient_clip_val: int | float | None
    """The value at which to clip gradients. Passing ``gradient_clip_val=None`` disables
    gradient clipping. If using Automatic Mixed Precision (AMP), the gradients will be unscaled before.
    Default: ``None``.
    """

    gradient_clip_algorithm: str | None
    """The gradient clipping algorithm to use. Pass ``gradient_clip_algorithm="value"``
    to clip by value, and ``gradient_clip_algorithm="norm"`` to clip by norm. By default it will
    be set to ``"norm"``.
    """

    deterministic: bool | Literal["warn"] | None
    """If ``True``, sets whether PyTorch operations must use deterministic algorithms.
    Set to ``"warn"`` to use deterministic algorithms whenever possible, throwing warnings on operations
    that don't support deterministic mode. If not set, defaults to ``False``. Default: ``None``.
    """

    benchmark: bool | None
    """The value (``True`` or ``False``) to set ``torch.backends.cudnn.benchmark`` to.
    The value for ``torch.backends.cudnn.benchmark`` set in the current session will be used
    (``False`` if not manually set). If :paramref:`~lightning.pytorch.trainer.trainer.Trainer.deterministic`
    is set to ``True``, this will default to ``False``. Override to manually set a different value.
    Default: ``None``.
    """

    inference_mode: bool
    """Whether to use :func:`torch.inference_mode` or :func:`torch.no_grad` during
    evaluation (``validate``/``test``/``predict``).
    """

    use_distributed_sampler: bool
    """Whether to wrap the DataLoader's sampler with
    :class:`torch.utils.data.DistributedSampler`. If not specified this is toggled automatically for
    strategies that require it. By default, it will add ``shuffle=True`` for the train sampler and
    ``shuffle=False`` for validation/test/predict samplers. If you want to disable this logic, you can pass
    ``False`` and add your own distributed sampler in the dataloader hooks. If ``True`` and a distributed
    sampler was already added, Lightning will not replace the existing one. For iterable-style datasets,
    we don't do this automatically.
    """

    profiler: Profiler | str | None
    """To profile individual steps during training and assist in identifying bottlenecks.
    Default: ``None``.
    """

    detect_anomaly: bool
    """Enable anomaly detection for the autograd engine.
    Default: ``False``.
    """

    barebones: bool
    """Whether to run in "barebones mode", where all features that may impact raw speed are
    disabled. This is meant for analyzing the Trainer overhead and is discouraged during regular training
    runs. The following features are deactivated:
    :paramref:`~lightning.pytorch.trainer.trainer.Trainer.enable_checkpointing`,
    :paramref:`~lightning.pytorch.trainer.trainer.Trainer.logger`,
    :paramref:`~lightning.pytorch.trainer.trainer.Trainer.enable_progress_bar`,
    :paramref:`~lightning.pytorch.trainer.trainer.Trainer.log_every_n_steps`,
    :paramref:`~lightning.pytorch.trainer.trainer.Trainer.enable_model_summary`,
    :paramref:`~lightning.pytorch.trainer.trainer.Trainer.num_sanity_val_steps`,
    :paramref:`~lightning.pytorch.trainer.trainer.Trainer.fast_dev_run`,
    :paramref:`~lightning.pytorch.trainer.trainer.Trainer.detect_anomaly`,
    :paramref:`~lightning.pytorch.trainer.trainer.Trainer.profiler`,
    :meth:`~lightning.pytorch.core.LightningModule.log`,
    :meth:`~lightning.pytorch.core.LightningModule.log_dict`.
    """

    plugins: _PLUGIN_INPUT | list[_PLUGIN_INPUT] | None
    """Plugins allow modification of core behavior like ddp and amp, and enable custom lightning plugins.
    Default: ``None``.
    """

    sync_batchnorm: bool
    """Synchronize batch norm layers between process groups/whole world.
    Default: ``False``.
    """

    reload_dataloaders_every_n_epochs: int
    """Set to a positive integer to reload dataloaders every n epochs.
    Default: ``0``.
    """

    default_root_dir: Path | None
    """Default path for logs and weights when no logger/ckpt_callback passed.
    Default: ``os.getcwd()``.
    Can be remote file paths such as `s3://mybucket/path` or 'hdfs://path/'
    """


class SanityCheckingConfig(C.Config):
    reduce_lr_on_plateau: Literal["disable", "error", "warn"] = "error"
    """
    If enabled, will do some sanity checks if the `ReduceLROnPlateau` scheduler is used:
        - If the `interval` is step, it makes sure that validation is called every `frequency` steps.
        - If the `interval` is epoch, it makes sure that validation is called every `frequency` epochs.
    Valid values are: "disable", "warn", "error".
    """


class TrainerConfig(C.Config):
    # region Active Run Configuration
    id: str = C.Field(default_factory=lambda: TrainerConfig.generate_id())
    """ID of the run."""
    name: list[str] = []
    """Run name in parts. Full name is constructed by joining the parts with spaces."""
    project: str | None = None
    """Project name."""
    tags: list[str] = []
    """Tags for the run."""
    notes: list[str] = []
    """Human readable notes for the run."""
    meta: dict[str, Any] = {}
    """Metadata information for the run. This is a dictionary that can be used to store any additional information
    about the run. It is not used by nshtrainer, but can be useful for logging or tracking purposes.
    """

    @property
    def full_name(self):
        return " ".join(self.name)

    debug: bool = False
    """Whether to run in debug mode. This will enable debug logging and enable debug code paths."""

    environment: Annotated[EnvironmentConfig, C.Field(repr=False)] = (
        EnvironmentConfig.empty()
    )
    """A snapshot of the current environment information (e.g. python version, slurm info, etc.). This is automatically populated by the run script."""

    directory: DirectoryConfig = DirectoryConfig()
    """Directory configuration options."""

    _rng: ClassVar[np.random.Generator | None] = None

    @classmethod
    def generate_id(cls, *, length: int = 8) -> str:
        """
        Generate a random ID of specified length.

        """
        if (rng := cls._rng) is None:
            rng = np.random.default_rng()

        alphabet = list(string.ascii_lowercase + string.digits)

        id = "".join(rng.choice(alphabet) for _ in range(length))
        return id

    @classmethod
    def set_seed(cls, seed: int | None = None) -> None:
        """
        Set the seed for the random number generator.

        Args:
            seed (int | None, optional): The seed value to set. If None, a seed based on the current time will be used. Defaults to None.

        Returns:
            None
        """
        if seed is None:
            seed = int(time.time() * 1000)
        log.critical(f"Seeding {cls.__name__} with seed {seed}")
        cls._rng = np.random.default_rng(seed)

    # endregion

    primary_metric: MetricConfig | None = None
    """Primary metric configuration options. This is used in the following ways:
    - To determine the best model checkpoint to save with the ModelCheckpoint callback.
    - To monitor the primary metric during training and stop training based on the `early_stopping` configuration.
    - For the ReduceLROnPlateau scheduler.
    """

    ckpt_path: Literal["none"] | str | Path | None = None
    """Path to a checkpoint to load and resume training from. If ``"none"``, will not load a checkpoint."""

    checkpoint_saving: CheckpointSavingConfig = CheckpointSavingConfig()
    """Checkpoint saving configuration options."""

    hf_hub: HuggingFaceHubConfig = HuggingFaceHubConfig()
    """Hugging Face Hub configuration options."""

    loggers: Sequence[LoggerConfig] | None = None
    """Loggers to use for experiment tracking."""

    def enabled_loggers(self) -> Sequence[LoggerConfig]:
        # Default loggers
        if (loggers := self.loggers) is None:
            loggers = [
                WandbLoggerConfig(),
                CSVLoggerConfig(),
                TensorboardLoggerConfig(),
            ]

        # In barebones mode, disable all loggers
        if self.barebones:
            loggers = []

        return loggers

    actsave_logger: ActSaveLoggerConfig | None = None
    """If enabled, will automatically save logged metrics using ActSave (if nshutils is installed)."""

    lr_monitor: LearningRateMonitorConfig | None = LearningRateMonitorConfig()
    """Learning rate monitoring configuration options."""

    log_epoch: LogEpochCallbackConfig | None = LogEpochCallbackConfig()
    """If enabled, will log the fractional epoch number to the logger."""

    gradient_clipping: GradientClippingConfig | None = None
    """Gradient clipping configuration, or None to disable gradient clipping."""

    log_norms: NormLoggingCallbackConfig | None = None
    """Norm logging configuration options."""

    deterministic: bool | Literal["warn"] | None = None
    """
    If ``True``, sets whether PyTorch operations must use deterministic algorithms.
        Set to ``"warn"`` to use deterministic algorithms whenever possible, throwing warnings on operations
        that don't support deterministic mode. If not set, defaults to ``False``. Default: ``None``.
    """

    reduce_lr_on_plateau_sanity_checking: RLPSanityChecksCallbackConfig | None = (
        RLPSanityChecksCallbackConfig()
    )
    """
    If enabled, will do some sanity checks if the `ReduceLROnPlateau` scheduler is used:
        - If the `interval` is step, it makes sure that validation is called every `frequency` steps.
        - If the `interval` is epoch, it makes sure that validation is called every `frequency` epochs.
    """

    early_stopping: EarlyStoppingCallbackConfig | None = None
    """Early stopping configuration options."""

    profiler: ProfilerConfig | None = None
    """
    To profile individual steps during training and assist in identifying bottlenecks.
        Default: ``None``.
    """

    callbacks: list[CallbackConfig] = []
    """Callbacks to use during training."""

    detect_anomaly: bool | None = None
    """Enable anomaly detection for the autograd engine.
    Default: ``False``.
    """

    plugins: list[PluginConfig] | None = None
    """
    Plugins allow modification of core behavior like ddp and amp, and enable custom lightning plugins.
        Default: ``None``.
    """

    auto_determine_num_nodes: bool = True
    """
    If enabled, will automatically determine the number of nodes for distributed training.

    This will only work on:
    - SLURM clusters
    - LSF clusters
    """

    fast_dev_run: int | bool = False
    """Runs n if set to ``n`` (int) else 1 if set to ``True`` batch(es)
    of train, val and test to find any bugs (ie: a sort of unit test).
    Default: ``False``.
    """

    barebones: bool = False
    """Whether to run in "barebones mode", where all features that may impact raw speed are
    disabled. This is meant for analyzing the Trainer overhead and is discouraged during regular training
    runs. The following features are deactivated:
    :paramref:`~lightning.pytorch.trainer.trainer.Trainer.enable_checkpointing`,
    :paramref:`~lightning.pytorch.trainer.trainer.Trainer.logger`,
    :paramref:`~lightning.pytorch.trainer.trainer.Trainer.enable_progress_bar`,
    :paramref:`~lightning.pytorch.trainer.trainer.Trainer.log_every_n_steps`,
    :paramref:`~lightning.pytorch.trainer.trainer.Trainer.enable_model_summary`,
    :paramref:`~lightning.pytorch.trainer.trainer.Trainer.num_sanity_val_steps`,
    :paramref:`~lightning.pytorch.trainer.trainer.Trainer.fast_dev_run`,
    :paramref:`~lightning.pytorch.trainer.trainer.Trainer.detect_anomaly`,
    :paramref:`~lightning.pytorch.trainer.trainer.Trainer.profiler`,
    :meth:`~lightning.pytorch.core.LightningModule.log`,
    :meth:`~lightning.pytorch.core.LightningModule.log_dict`.
    """

    experimental_barebones_profiler_enabled: bool = False
    """If enabled, will enable the profiler in barebones mode.
    The implementation of this is very hacky and may not work as expected."""

    experimental_barebones_progress_bar_enabled: bool = False
    """If enabled, will enable the progress bar in barebones mode.
    The implementation of this is very hacky and may not work as expected."""

    precision: (
        Literal[
            "64-true",
            "32-true",
            "fp16-mixed",
            "bf16-mixed",
            "16-mixed-auto",
        ]
        | None
    ) = None
    """
    Training precision. Can be one of:
        - "64-true": Double precision (64-bit).
        - "32-true": Full precision (32-bit).
        - "fp16-mixed": Float16 mixed precision.
        - "bf16-mixed": BFloat16 mixed precision.
        - "16-mixed-auto": Automatic 16-bit: Uses bfloat16 if available, otherwise float16.
    """

    max_epochs: int | None = None
    """Stop training once this number of epochs is reached. Disabled by default (None).
    If both max_epochs and max_steps are not specified, defaults to ``max_epochs = 1000``.
    To enable infinite training, set ``max_epochs = -1``.
    """

    min_epochs: int | None = None
    """Force training for at least these many epochs. Disabled by default (None).
    """

    max_steps: int = -1
    """Stop training after this number of steps. Disabled by default (-1). If ``max_steps = -1``
    and ``max_epochs = None``, will default to ``max_epochs = 1000``. To enable infinite training, set
    ``max_epochs`` to ``-1``.
    """

    min_steps: int | None = None
    """Force training for at least these number of steps. Disabled by default (``None``).
    """

    max_time: str | timedelta | dict[str, int] | None = None
    """Stop training after this amount of time has passed. Disabled by default (``None``).
    The time duration can be specified in the format DD:HH:MM:SS (days, hours, minutes seconds), as a
    :class:`datetime.timedelta`, or a dictionary with keys that will be passed to
    :class:`datetime.timedelta`.
    """

    limit_train_batches: int | float | None = None
    """How much of training dataset to check (float = fraction, int = num_batches).
    Default: ``1.0``.
    """

    limit_val_batches: int | float | None = None
    """How much of validation dataset to check (float = fraction, int = num_batches).
    Default: ``1.0``.
    """

    limit_test_batches: int | float | None = None
    """How much of test dataset to check (float = fraction, int = num_batches).
    Default: ``1.0``.
    """

    limit_predict_batches: int | float | None = None
    """How much of prediction dataset to check (float = fraction, int = num_batches).
    Default: ``1.0``.
    """

    overfit_batches: int | float = 0.0
    """Overfit a fraction of training/validation data (float) or a set number of batches (int).
    Default: ``0.0``.
    """

    val_check_interval: int | float | None = None
    """How often to check the validation set. Pass a ``float`` in the range [0.0, 1.0] to check
    after a fraction of the training epoch. Pass an ``int`` to check after a fixed number of training
    batches. An ``int`` value can only be higher than the number of training batches when
    ``check_val_every_n_epoch=None``, which validates after every ``N`` training batches
    across epochs or during iteration-based training.
    Default: ``1.0``.
    """

    check_val_every_n_epoch: int | None = 1
    """Perform a validation loop every after every `N` training epochs. If ``None``,
    validation will be done solely based on the number of training batches, requiring ``val_check_interval``
    to be an integer value.
    Default: ``1``.
    """

    num_sanity_val_steps: int | None = None
    """Sanity check runs n validation batches before starting the training routine.
    Set it to `-1` to run all batches in all validation dataloaders.
    Default: ``2``.
    """

    log_every_n_steps: int | None = None
    """How often to log within steps.
    Default: ``50``.
    """

    inference_mode: bool = True
    """Whether to use :func:`torch.inference_mode` (if `True`) or :func:`torch.no_grad` (if `False`) during evaluation (``validate``/``test``/``predict``).
    Default: ``True``.
    """

    use_distributed_sampler: bool | None = None
    """Whether to wrap the DataLoader's sampler with
    :class:`torch.utils.data.DistributedSampler`. If not specified this is toggled automatically for
    strategies that require it. By default, it will add ``shuffle=True`` for the train sampler and
    ``shuffle=False`` for validation/test/predict samplers. If you want to disable this logic, you can pass
    ``False`` and add your own distributed sampler in the dataloader hooks. If ``True`` and a distributed
    sampler was already added, Lightning will not replace the existing one. For iterable-style datasets,
    we don't do this automatically.
    Default: ``True``.
    """

    accelerator: AcceleratorConfig | AcceleratorLiteral | None = None
    """Supports passing different accelerator types ("cpu", "gpu", "tpu", "ipu", "hpu", "mps", "auto")
    as well as custom accelerator instances.
    Default: ``"auto"``.
    """

    strategy: StrategyConfig | StrategyLiteral | None = None
    """Supports different training strategies with aliases as well custom strategies.
    Default: ``"auto"``.
    """

    devices: tuple[int, ...] | Sequence[int] | Literal["auto", "all"] | None = None
    """The devices to use. Can be set to a sequence of device indices, "all" to indicate all available devices should be used, or ``"auto"`` for
    automatic selection based on the chosen accelerator. Default: ``"auto"``.
    """

    shared_parameters: SharedParametersCallbackConfig | None = None
    """If enabled, the model supports scaling the gradients of shared parameters that
    are registered in the self.shared_parameters list. This is useful for models that
    share parameters across multiple modules (e.g., in a GPT model) and want to
    downscale the gradients of these parameters to avoid overfitting.
    """

    auto_set_default_root_dir: bool = True
    """If enabled, will automatically set the default root dir to [cwd/lightning_logs/<id>/]. There is basically no reason to disable this."""
    save_checkpoint_metadata: bool = True
    """If enabled, will save additional metadata whenever a checkpoint is saved."""
    auto_set_debug_flag: DebugFlagCallbackConfig | None = DebugFlagCallbackConfig()
    """If enabled, will automatically set the debug flag to True if:
    - The trainer is running in fast_dev_run mode.
    - The trainer is running a sanity check (which happens before starting the training routine).
    """
    auto_validate_metrics: MetricValidationCallbackConfig | None = None
    """If enabled, will automatically validate the metrics before starting the training routine."""

    lightning_kwargs: LightningTrainerKwargs = LightningTrainerKwargs()
    """
    Additional keyword arguments to pass to the Lightning `pl.Trainer` constructor.

    Please refer to the Lightning documentation for a list of valid keyword arguments.
    """

    additional_lightning_kwargs: dict[str, Any] = {}
    """
    Additional keyword arguments to pass to the Lightning `pl.Trainer` constructor.

    This is essentially a non-type-checked version of `lightning_kwargs`.
    """

    set_float32_matmul_precision: Literal["medium", "high", "highest"] | None = None
    """If enabled, will set the torch float32 matmul precision to the specified value. Useful for faster training on Ampere+ GPUs."""

    @property
    def wandb_logger(self):
        return next(
            (
                logger
                for logger in self.enabled_loggers()
                if isinstance(logger, WandbLoggerConfig)
            ),
            None,
        )

    @property
    def csv_logger(self):
        return next(
            (
                logger
                for logger in self.enabled_loggers()
                if isinstance(logger, CSVLoggerConfig)
            ),
            None,
        )

    @property
    def tensorboard_logger(self):
        return next(
            (
                logger
                for logger in self.enabled_loggers()
                if isinstance(logger, TensorboardLoggerConfig)
            ),
            None,
        )

    def _nshtrainer_all_callback_configs(self) -> Iterable[CallbackConfigBase | None]:
        yield self.directory.setup_callback
        yield self.early_stopping
        yield self.checkpoint_saving
        yield self.lr_monitor
        yield from (
            logger_config
            for logger_config in self.enabled_loggers()
            if logger_config is not None
            and isinstance(logger_config, CallbackConfigBase)
        )
        yield self.log_epoch
        yield self.log_norms
        yield self.hf_hub
        yield self.shared_parameters
        yield self.reduce_lr_on_plateau_sanity_checking
        yield self.auto_set_debug_flag
        yield self.auto_validate_metrics
        yield from self.callbacks

    def _nshtrainer_all_logger_configs(self) -> Iterable[LoggerConfigBase | None]:
        # Disable all loggers if barebones mode is enabled
        if self.barebones:
            return

        yield from self.enabled_loggers()
        yield self.actsave_logger

    def _nshtrainer_validate_before_run(self):
        # shared_parameters is not supported under barebones mode
        if self.barebones and self.shared_parameters:
            raise ValueError("shared_parameters is not supported under barebones mode")

    # region Helper Methods
    def fast_dev_run_(self, value: int | bool = True, /):
        """
        Enables fast_dev_run mode for the trainer.
        This will run the training loop for a specified number of batches,
        if an integer is provided, or for a single batch if True is provided.
        """
        self.fast_dev_run = value
        return self

    def with_fast_dev_run(self, value: int | bool = True, /):
        """
        Enables fast_dev_run mode for the trainer.
        This will run the training loop for a specified number of batches,
        if an integer is provided, or for a single batch if True is provided.
        """
        return copy.deepcopy(self).fast_dev_run_(value)

    def project_root_(self, project_root: str | Path | os.PathLike):
        """
        Set the project root directory for the trainer.

        Args:
            project_root (Path): The base directory to use.

        Returns:
            self: The current instance of the class.
        """
        self.directory.project_root = Path(project_root)
        return self

    def with_project_root(self, project_root: str | Path | os.PathLike):
        """
        Set the project root directory for the trainer.

        Args:
            project_root (Path): The base directory to use.

        Returns:
            self: The current instance of the class.
        """
        return copy.deepcopy(self).project_root_(project_root)

    def reset_run(
        self,
        *,
        id: bool = True,
        basic: bool = True,
        project_root: bool = True,
        environment: bool = True,
    ):
        """
        Reset the configuration object to its initial state.

        Parameters:
        - id (bool): If True, generate a new ID for the configuration object.
        - basic (bool): If True, reset basic attributes like name, project, tags, and notes.
        - project_root (bool): If True, reset the directory configuration to its initial state.
        - environment (bool): If True, reset the environment configuration to its initial state.
        - meta (bool): If True, reset the meta dictionary to an empty dictionary.

        Returns:
        - self: The updated configuration object.

        """
        config = copy.deepcopy(self)

        if id:
            config.id = config.generate_id()

        if basic:
            config.name = []
            config.project = None
            config.tags = []
            config.notes = []

        if project_root:
            config.directory = DirectoryConfig()

        if environment:
            config.environment = EnvironmentConfig.empty()

        return config

    # endregion



from __future__ import annotations

from typing import Annotated

from typing_extensions import TypeAliasType

from .actsave import ActSaveLoggerConfig as ActSaveLoggerConfig
from .base import LoggerConfigBase as LoggerConfigBase
from .base import logger_registry as logger_registry
from .csv import CSVLoggerConfig as CSVLoggerConfig
from .tensorboard import TensorboardLoggerConfig as TensorboardLoggerConfig
from .wandb import WandbLoggerConfig as WandbLoggerConfig

LoggerConfig = TypeAliasType(
    "LoggerConfig",
    Annotated[LoggerConfigBase, logger_registry.DynamicResolution()],
)



from __future__ import annotations

__codegen__ = True

from nshtrainer.lr_scheduler.linear_warmup_cosine import (
    DurationConfig as DurationConfig,
)
from nshtrainer.lr_scheduler.linear_warmup_cosine import (
    LinearWarmupCosineDecayLRSchedulerConfig as LinearWarmupCosineDecayLRSchedulerConfig,
)
from nshtrainer.lr_scheduler.linear_warmup_cosine import (
    LRSchedulerConfigBase as LRSchedulerConfigBase,
)
from nshtrainer.lr_scheduler.linear_warmup_cosine import (
    lr_scheduler_registry as lr_scheduler_registry,
)

__all__ = [
    "DurationConfig",
    "LRSchedulerConfigBase",
    "LinearWarmupCosineDecayLRSchedulerConfig",
    "lr_scheduler_registry",
]



from __future__ import annotations

__codegen__ = True

from nshtrainer.trainer.plugin.precision import (
    BitsandbytesPluginConfig as BitsandbytesPluginConfig,
)
from nshtrainer.trainer.plugin.precision import (
    DeepSpeedPluginConfig as DeepSpeedPluginConfig,
)
from nshtrainer.trainer.plugin.precision import (
    DoublePrecisionPluginConfig as DoublePrecisionPluginConfig,
)
from nshtrainer.trainer.plugin.precision import DTypeConfig as DTypeConfig
from nshtrainer.trainer.plugin.precision import (
    FSDPPrecisionPluginConfig as FSDPPrecisionPluginConfig,
)
from nshtrainer.trainer.plugin.precision import (
    HalfPrecisionPluginConfig as HalfPrecisionPluginConfig,
)
from nshtrainer.trainer.plugin.precision import (
    MixedPrecisionPluginConfig as MixedPrecisionPluginConfig,
)
from nshtrainer.trainer.plugin.precision import PluginConfigBase as PluginConfigBase
from nshtrainer.trainer.plugin.precision import (
    TransformerEnginePluginConfig as TransformerEnginePluginConfig,
)
from nshtrainer.trainer.plugin.precision import XLAPluginConfig as XLAPluginConfig
from nshtrainer.trainer.plugin.precision import plugin_registry as plugin_registry

__all__ = [
    "BitsandbytesPluginConfig",
    "DTypeConfig",
    "DeepSpeedPluginConfig",
    "DoublePrecisionPluginConfig",
    "FSDPPrecisionPluginConfig",
    "HalfPrecisionPluginConfig",
    "MixedPrecisionPluginConfig",
    "PluginConfigBase",
    "TransformerEnginePluginConfig",
    "XLAPluginConfig",
    "plugin_registry",
]



from __future__ import annotations

__codegen__ = True

from nshtrainer.trainer.plugin.environment import (
    KubeflowEnvironmentPlugin as KubeflowEnvironmentPlugin,
)
from nshtrainer.trainer.plugin.environment import (
    LightningEnvironmentPlugin as LightningEnvironmentPlugin,
)
from nshtrainer.trainer.plugin.environment import (
    LSFEnvironmentPlugin as LSFEnvironmentPlugin,
)
from nshtrainer.trainer.plugin.environment import (
    MPIEnvironmentPlugin as MPIEnvironmentPlugin,
)
from nshtrainer.trainer.plugin.environment import PluginConfigBase as PluginConfigBase
from nshtrainer.trainer.plugin.environment import (
    SLURMEnvironmentPlugin as SLURMEnvironmentPlugin,
)
from nshtrainer.trainer.plugin.environment import (
    TorchElasticEnvironmentPlugin as TorchElasticEnvironmentPlugin,
)
from nshtrainer.trainer.plugin.environment import (
    XLAEnvironmentPlugin as XLAEnvironmentPlugin,
)
from nshtrainer.trainer.plugin.environment import plugin_registry as plugin_registry

__all__ = [
    "KubeflowEnvironmentPlugin",
    "LSFEnvironmentPlugin",
    "LightningEnvironmentPlugin",
    "MPIEnvironmentPlugin",
    "PluginConfigBase",
    "SLURMEnvironmentPlugin",
    "TorchElasticEnvironmentPlugin",
    "XLAEnvironmentPlugin",
    "plugin_registry",
]



from __future__ import annotations

from collections.abc import Callable
from typing import Literal

from lightning.pytorch import LightningModule, Trainer
from lightning.pytorch.callbacks import Callback
from typing_extensions import override

Split = Literal["train", "val", "test", "predict"]


def _check_step(step: int, interval: int, skip_first: bool = False):
    if step % interval != 0:
        return False
    if skip_first and step == 0:
        return False
    return True


class StepIntervalCallback(Callback):
    def __init__(
        self,
        function: Callable[[Trainer, LightningModule], None],
        *,
        interval: int,
        skip_first: bool = False,
        splits: list[Split] = ["train", "val", "test", "predict"],
    ):
        super().__init__()

        self.function = function
        self.interval = interval
        self.skip_first = skip_first
        self.splits = set(splits)

    @override
    def on_train_batch_start(
        self,
        trainer,
        pl_module,
        batch,
        batch_idx,
        dataloader_idx=1,
    ):
        if (
            not _check_step(
                trainer.global_step,
                self.interval,
                skip_first=self.skip_first,
            )
            or "train" not in self.splits
        ):
            return
        self.function(trainer, pl_module)

    @override
    def on_validation_batch_start(
        self,
        trainer,
        pl_module,
        batch,
        batch_idx,
        dataloader_idx=1,
    ):
        if (
            not _check_step(
                trainer.global_step,
                self.interval,
                skip_first=self.skip_first,
            )
            or "val" not in self.splits
        ):
            return
        self.function(trainer, pl_module)

    @override
    def on_test_batch_start(
        self,
        trainer,
        pl_module,
        batch,
        batch_idx,
        dataloader_idx=1,
    ):
        if (
            not _check_step(
                trainer.global_step,
                self.interval,
                skip_first=self.skip_first,
            )
            or "test" not in self.splits
        ):
            return
        self.function(trainer, pl_module)

    @override
    def on_predict_batch_start(
        self,
        trainer,
        pl_module,
        batch,
        batch_idx,
        dataloader_idx=1,
    ):
        if (
            not _check_step(
                trainer.global_step,
                self.interval,
                skip_first=self.skip_first,
            )
            or "predict" not in self.splits
        ):
            return
        self.function(trainer, pl_module)


class EpochIntervalCallback(Callback):
    def __init__(
        self,
        function: Callable[[Trainer, LightningModule], None],
        *,
        interval: int,
        skip_first: bool = False,
        splits: list[Split] = ["train", "val", "test", "predict"],
    ):
        super().__init__()

        self.function = function
        self.interval = interval
        self.skip_first = skip_first
        self.splits = set(splits)

    @override
    def on_train_epoch_start(self, trainer, pl_module):
        if (
            not _check_step(
                trainer.current_epoch,
                self.interval,
                skip_first=self.skip_first,
            )
            or "train" not in self.splits
        ):
            return
        self.function(trainer, pl_module)

    @override
    def on_validation_epoch_start(self, trainer, pl_module):
        if (
            not _check_step(
                trainer.current_epoch,
                self.interval,
                skip_first=self.skip_first,
            )
            or "val" not in self.splits
        ):
            return
        self.function(trainer, pl_module)

    @override
    def on_test_epoch_start(self, trainer, pl_module):
        if (
            not _check_step(
                trainer.current_epoch,
                self.interval,
                skip_first=self.skip_first,
            )
            or "test" not in self.splits
        ):
            return
        self.function(trainer, pl_module)

    @override
    def on_predict_epoch_start(self, trainer, pl_module):
        if (
            not _check_step(
                trainer.current_epoch,
                self.interval,
                skip_first=self.skip_first,
            )
            or "predict" not in self.splits
        ):
            return
        self.function(trainer, pl_module)


class IntervalCallback(Callback):
    def __init__(
        self,
        function: Callable[[Trainer, LightningModule], None],
        *,
        step_interval: int | None = None,
        epoch_interval: int | None = None,
        skip_first: bool = False,
        splits: list[Split] = ["train", "val", "test", "predict"],
    ):
        super().__init__()

        self.callback = None

        if step_interval is not None:
            self.callback = StepIntervalCallback(
                function,
                interval=step_interval,
                splits=splits,
                skip_first=skip_first,
            )
        elif epoch_interval is not None:
            self.callback = EpochIntervalCallback(
                function,
                interval=epoch_interval,
                splits=splits,
                skip_first=skip_first,
            )
        else:
            raise ValueError("Either step_interval or epoch_interval must be specified")

    @override
    def on_train_batch_start(
        self,
        trainer,
        pl_module,
        batch,
        batch_idx,
        dataloader_idx=1,
    ):
        if not isinstance(self.callback, StepIntervalCallback):
            return

        self.callback.on_train_batch_start(
            trainer,
            pl_module,
            batch,
            batch_idx,
            dataloader_idx=1,
        )

    @override
    def on_validation_batch_start(
        self,
        trainer,
        pl_module,
        batch,
        batch_idx,
        dataloader_idx=1,
    ):
        if not isinstance(self.callback, StepIntervalCallback):
            return

        self.callback.on_validation_batch_start(
            trainer,
            pl_module,
            batch,
            batch_idx,
            dataloader_idx=1,
        )

    @override
    def on_test_batch_start(
        self,
        trainer,
        pl_module,
        batch,
        batch_idx,
        dataloader_idx=1,
    ):
        if not isinstance(self.callback, StepIntervalCallback):
            return

        self.callback.on_test_batch_start(
            trainer,
            pl_module,
            batch,
            batch_idx,
            dataloader_idx=1,
        )

    @override
    def on_predict_batch_start(
        self,
        trainer,
        pl_module,
        batch,
        batch_idx,
        dataloader_idx=1,
    ):
        if not isinstance(self.callback, StepIntervalCallback):
            return

        self.callback.on_predict_batch_start(
            trainer,
            pl_module,
            batch,
            batch_idx,
            dataloader_idx=1,
        )

    @override
    def on_train_epoch_start(self, trainer, pl_module):
        if not isinstance(self.callback, EpochIntervalCallback):
            return

        self.callback.on_train_epoch_start(trainer, pl_module)

    @override
    def on_validation_epoch_start(self, trainer, pl_module):
        if not isinstance(self.callback, EpochIntervalCallback):
            return

        self.callback.on_validation_epoch_start(trainer, pl_module)

    @override
    def on_test_epoch_start(self, trainer, pl_module):
        if not isinstance(self.callback, EpochIntervalCallback):
            return

        self.callback.on_test_epoch_start(trainer, pl_module)

    @override
    def on_predict_epoch_start(self, trainer, pl_module):
        if not isinstance(self.callback, EpochIntervalCallback):
            return

        self.callback.on_predict_epoch_start(trainer, pl_module)



from __future__ import annotations

from typing import TYPE_CHECKING

from typing_extensions import TypeVar

TBase = TypeVar("TBase", infer_variance=True)


def mixin_base_type(base_class: type[TBase]) -> type[TBase]:
    """
    Useful function to make mixins with baseclass typehint

    ```
    class ReadonlyMixin(mixin_base_type(BaseAdmin))):
        ...
    ```
    """
    if TYPE_CHECKING:
        return base_class
    return object



from __future__ import annotations

from typing import TYPE_CHECKING, Literal

import nshconfig as C
import torch
from typing_extensions import TypeAliasType, assert_never

from ..bf16 import is_bf16_supported_no_emulation

if TYPE_CHECKING:
    from ...trainer._config import TrainerConfig

DTypeName = TypeAliasType(
    "DTypeName",
    Literal[
        "float32",
        "float",
        "float64",
        "double",
        "float16",
        "bfloat16",
        "float8_e4m3fn",
        "float8_e4m3fnuz",
        "float8_e5m2",
        "float8_e5m2fnuz",
        "half",
        "uint8",
        "uint16",
        "uint32",
        "uint64",
        "int8",
        "int16",
        "short",
        "int32",
        "int",
        "int64",
        "long",
        "complex32",
        "complex64",
        "chalf",
        "cfloat",
        "complex128",
        "cdouble",
        "quint8",
        "qint8",
        "qint32",
        "bool",
        "quint4x2",
        "quint2x4",
        "bits1x8",
        "bits2x4",
        "bits4x2",
        "bits8",
        "bits16",
    ],
)


class DTypeConfig(C.Config):
    name: DTypeName
    """The name of the dtype."""

    @classmethod
    def from_trainer_config(cls, trainer_config: TrainerConfig):
        if (precision := trainer_config.precision) is None:
            precision = "32-true"

        match precision:
            case "16-mixed-auto":
                return (
                    cls(name="bfloat16")
                    if is_bf16_supported_no_emulation()
                    else cls(name="float16")
                )
            case "fp16-mixed":
                return cls(name="float16")
            case "bf16-mixed":
                return cls(name="bfloat16")
            case "32-true":
                return cls(name="float32")
            case "64-true":
                return cls(name="float64")
            case _:
                assert_never(trainer_config.precision)

    @property
    def torch_dtype(self):
        if ((dtype := getattr(torch, self.name, None)) is None) or not isinstance(
            dtype, torch.dtype
        ):
            raise ValueError(f"Unknown dtype {self.name}")

        return dtype



from __future__ import annotations

__codegen__ = True

from nshtrainer.nn.mlp import MLPConfig as MLPConfig
from nshtrainer.nn.mlp import NonlinearityConfig as NonlinearityConfig
from nshtrainer.nn.mlp import NonlinearityConfigBase as NonlinearityConfigBase

__all__ = [
    "MLPConfig",
    "NonlinearityConfig",
    "NonlinearityConfigBase",
]



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Literal

import nshconfig as C
from lightning.pytorch.strategies.strategy import Strategy
from typing_extensions import TypeAliasType

if TYPE_CHECKING:
    from ._config import TrainerConfig

StrategyLiteral = TypeAliasType(
    "StrategyLiteral",
    Literal[
        "auto",
        "ddp",
        "ddp_find_unused_parameters_false",
        "ddp_find_unused_parameters_true",
        "ddp_spawn",
        "ddp_spawn_find_unused_parameters_false",
        "ddp_spawn_find_unused_parameters_true",
        "ddp_fork",
        "ddp_fork_find_unused_parameters_false",
        "ddp_fork_find_unused_parameters_true",
        "ddp_notebook",
        "dp",
        "deepspeed",
        "deepspeed_stage_1",
        "deepspeed_stage_1_offload",
        "deepspeed_stage_2",
        "deepspeed_stage_2_offload",
        "deepspeed_stage_3",
        "deepspeed_stage_3_offload",
        "deepspeed_stage_3_offload_nvme",
        "fsdp",
        "fsdp_cpu_offload",
        "single_xla",
        "xla_fsdp",
        "xla",
        "single_tpu",
    ],
)


class StrategyConfigBase(C.Config, ABC):
    @abstractmethod
    def create_strategy(self, trainer_config: "TrainerConfig") -> Strategy: ...


StrategyConfig = TypeAliasType("StrategyConfig", StrategyConfigBase)



from __future__ import annotations

__codegen__ = True

from nshtrainer.profiler import AdvancedProfilerConfig as AdvancedProfilerConfig
from nshtrainer.profiler import BaseProfilerConfig as BaseProfilerConfig
from nshtrainer.profiler import ProfilerConfig as ProfilerConfig
from nshtrainer.profiler import PyTorchProfilerConfig as PyTorchProfilerConfig
from nshtrainer.profiler import SimpleProfilerConfig as SimpleProfilerConfig

from . import _base as _base
from . import advanced as advanced
from . import pytorch as pytorch
from . import simple as simple

__all__ = [
    "AdvancedProfilerConfig",
    "BaseProfilerConfig",
    "ProfilerConfig",
    "PyTorchProfilerConfig",
    "SimpleProfilerConfig",
    "_base",
    "advanced",
    "pytorch",
    "simple",
]



from __future__ import annotations

import logging
import time
from datetime import timedelta
from pathlib import Path
from typing import Any, Literal

from lightning.pytorch import LightningModule, Trainer
from typing_extensions import final, override

from ..._checkpoint.metadata import CheckpointMetadata
from ..base import callback_registry
from ._base import BaseCheckpointCallbackConfig, CheckpointBase

log = logging.getLogger(__name__)


@final
@callback_registry.register
class LastCheckpointCallbackConfig(BaseCheckpointCallbackConfig):
    name: Literal["last_checkpoint"] = "last_checkpoint"

    save_on_time_interval: timedelta | None = None
    """Save a checkpoint every `save_on_time_interval` seconds. If `None`, this feature is disabled."""

    @override
    def create_checkpoint(self, trainer_config, dirpath):
        return LastCheckpointCallback(self, dirpath)


@final
class LastCheckpointCallback(CheckpointBase[LastCheckpointCallbackConfig]):
    def __init__(self, config: LastCheckpointCallbackConfig, dirpath: Path):
        super().__init__(config, dirpath)
        self.start_time = time.time()
        self.last_checkpoint_time = self.start_time

    @override
    def name(self):
        return "last"

    @override
    def default_filename(self):
        return "epoch{epoch}-step{step}-duration{train_duration}"

    @override
    def topk_sort_key(self, metadata: CheckpointMetadata):
        return metadata.checkpoint_timestamp

    @override
    def topk_sort_reverse(self):
        return True

    def _local_should_checkpoint(self) -> bool:
        if (interval := self.config.save_on_time_interval) is None:
            return False

        current_time = time.time()
        elapsed_time = current_time - self.last_checkpoint_time
        return elapsed_time >= interval.total_seconds()

    def _should_checkpoint(self, trainer: Trainer):
        if self.config.save_on_time_interval is None:
            return False
        return trainer.strategy.broadcast(self._local_should_checkpoint(), src=0)

    def _format_duration(self, seconds: float) -> str:
        """Format duration in seconds to a human-readable string."""
        td = timedelta(seconds=int(seconds))
        days = td.days
        hours, remainder = divmod(td.seconds, 3600)
        minutes, seconds = divmod(remainder, 60)

        parts = []
        if days > 0:
            parts.append(f"{days}d")
        if hours > 0:
            parts.append(f"{hours}h")
        if minutes > 0:
            parts.append(f"{minutes}m")
        if seconds > 0 or not parts:
            parts.append(f"{seconds}s")

        return "_".join(parts)

    @override
    def current_metrics(self, trainer: Trainer) -> dict[str, Any]:
        metrics = super().current_metrics(trainer)
        train_duration = time.time() - self.start_time
        metrics["train_duration"] = self._format_duration(train_duration)
        return metrics

    @override
    def on_train_batch_end(
        self,
        trainer: Trainer,
        pl_module: LightningModule,
        *args,
        **kwargs,
    ):
        if not self._should_checkpoint(trainer):
            return
        self.save_checkpoints(trainer)

    @override
    def on_validation_epoch_end(self, trainer: Trainer, pl_module: LightningModule):
        self.save_checkpoints(trainer)

    @override
    def save_checkpoints(self, trainer):
        super().save_checkpoints(trainer)

        if self.config.save_on_time_interval is not None:
            self.last_checkpoint_time = time.time()



from __future__ import annotations

import dataclasses
from collections import deque
from collections.abc import Callable, Generator, Mapping
from contextlib import contextmanager
from typing import Any, ClassVar

import torchmetrics
from lightning.pytorch import LightningModule
from lightning.pytorch.utilities.types import _METRIC
from lightning_utilities.core.rank_zero import rank_zero_warn
from typing_extensions import override

from ...util.typing_utils import mixin_base_type


@dataclasses.dataclass(frozen=True, kw_only=True)
class _LogContextKwargs:
    __ignore_fields__: ClassVar[set[str]] = {"prefix", "disabled"}

    prefix: str | None = None
    disabled: bool | None = None
    prog_bar: bool | None = None
    logger: bool | None = None
    on_step: bool | None = None
    on_epoch: bool | None = None
    reduce_fx: str | Callable | None = None
    enable_graph: bool | None = None
    sync_dist: bool | None = None
    sync_dist_group: Any | None = None
    add_dataloader_idx: bool | None = None
    batch_size: int | None = None
    rank_zero_only: bool | None = None

    def to_dict(self):
        d = dataclasses.asdict(self)
        for field in self.__ignore_fields__:
            d.pop(field, None)

        # Pop all None values
        for k in list(d.keys()):
            if d[k] is None:
                d.pop(k)

        return d


class LoggerLightningModuleMixin(mixin_base_type(LightningModule)):
    @override
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._logger_prefix_stack = deque[_LogContextKwargs]()

    @property
    def logging_enabled(self) -> bool:
        # Logging is disabled in barebones mode.
        if (trainer := self._trainer) is not None and trainer.barebones:
            # Warn the user once that logging is disabled in barebones mode.
            if not hasattr(self, "_barebones_logging_warned"):
                rank_zero_warn(
                    "Logging is disabled in barebones mode. "
                    "This is to reduce the overhead of logging in barebones mode. "
                    "If you want to enable logging, set `barebones=False` in the Trainer.",
                )
                self._barebones_logging_warned = True
            return False

        # If no loggers are registered, then logging is disabled.
        if not self.logger:
            return False

        # Check if the topmost non-null context is disabled
        for context in reversed(self._logger_prefix_stack):
            if context.disabled is not None:
                return not context.disabled

        # Otherwise, logging is enabled.
        return True

    @contextmanager
    def log_context(
        self,
        prefix: str | None = None,
        disabled: bool | None = None,
        prog_bar: bool | None = None,
        logger: bool | None = None,
        on_step: bool | None = None,
        on_epoch: bool | None = None,
        reduce_fx: str | Callable | None = None,
        enable_graph: bool | None = None,
        sync_dist: bool | None = None,
        sync_dist_group: Any | None = None,
        add_dataloader_idx: bool | None = None,
        batch_size: int | None = None,
        rank_zero_only: bool | None = None,
    ) -> Generator[None, None, None]:
        self._logger_prefix_stack.append(
            _LogContextKwargs(
                prefix=prefix,
                disabled=disabled,
                prog_bar=prog_bar,
                logger=logger,
                on_step=on_step,
                on_epoch=on_epoch,
                reduce_fx=reduce_fx,
                enable_graph=enable_graph,
                sync_dist=sync_dist,
                sync_dist_group=sync_dist_group,
                add_dataloader_idx=add_dataloader_idx,
                batch_size=batch_size,
                rank_zero_only=rank_zero_only,
            )
        )
        try:
            yield
        finally:
            _ = self._logger_prefix_stack.pop()

    def _make_prefix_and_kwargs_dict(self, kwargs: _LogContextKwargs):
        prefix = "".join(c.prefix for c in self._logger_prefix_stack if c.prefix)

        fn_kwargs: dict[str, Any] = {}
        for c in self._logger_prefix_stack:
            fn_kwargs.update(c.to_dict())

        fn_kwargs.update(kwargs.to_dict())
        return prefix, fn_kwargs

    @override
    def log(
        self,
        name: str,
        value: _METRIC,
        prog_bar: bool | None = None,
        logger: bool | None = None,
        on_step: bool | None = None,
        on_epoch: bool | None = None,
        reduce_fx: str | Callable | None = None,
        enable_graph: bool | None = None,
        sync_dist: bool | None = None,
        sync_dist_group: Any | None = None,
        add_dataloader_idx: bool | None = None,
        batch_size: int | None = None,
        metric_attribute: str | None = None,
        rank_zero_only: bool | None = None,
    ) -> None:
        """Log a key, value pair.

        Example::

            self.log('train_loss', loss)

        The default behavior per hook is documented here: :ref:`extensions/logging:Automatic Logging`.

        Args:
            name: key to log. Must be identical across all processes if using DDP or any other distributed strategy.
            value: value to log. Can be a ``float``, ``Tensor``, or a ``Metric``.
            prog_bar: if ``True`` logs to the progress bar.
            logger: if ``True`` logs to the logger.
            on_step: if ``True`` logs at this step. The default value is determined by the hook.
                See :ref:`extensions/logging:Automatic Logging` for details.
            on_epoch: if ``True`` logs epoch accumulated metrics. The default value is determined by the hook.
                See :ref:`extensions/logging:Automatic Logging` for details.
            reduce_fx: reduction function over step values for end of epoch. :meth:`torch.mean` by default.
            enable_graph: if ``True``, will not auto detach the graph.
            sync_dist: if ``True``, reduces the metric across devices. Use with care as this may lead to a significant
                communication overhead.
            sync_dist_group: the DDP group to sync across.
            add_dataloader_idx: if ``True``, appends the index of the current dataloader to
                the name (when using multiple dataloaders). If False, user needs to give unique names for
                each dataloader to not mix the values.
            batch_size: Current batch_size. This will be directly inferred from the loaded batch,
                but for some data structures you might need to explicitly provide it.
            metric_attribute: To restore the metric state, Lightning requires the reference of the
                :class:`torchmetrics.Metric` in your model. This is found automatically if it is a model attribute.
            rank_zero_only: Tells Lightning if you are calling ``self.log`` from every process (default) or only from
                rank 0. If ``True``, you won't be able to use this metric as a monitor in callbacks
                (e.g., early stopping). Warning: Improper use can lead to deadlocks! See
                :ref:`Advanced Logging <visualize/logging_advanced:rank_zero_only>` for more details.

        """
        # If logging is disabled, then do nothing.
        if not self.logging_enabled:
            return

        prefix, fn_kwargs = self._make_prefix_and_kwargs_dict(
            _LogContextKwargs(
                prog_bar=prog_bar,
                logger=logger,
                on_step=on_step,
                on_epoch=on_epoch,
                reduce_fx=reduce_fx,
                enable_graph=enable_graph,
                sync_dist=sync_dist,
                sync_dist_group=sync_dist_group,
                add_dataloader_idx=add_dataloader_idx,
                batch_size=batch_size,
                rank_zero_only=rank_zero_only,
            )
        )
        name = f"{prefix}{name}"
        return super().log(name, value, metric_attribute=metric_attribute, **fn_kwargs)

    def log_dict(
        self,
        dictionary: Mapping[str, _METRIC] | torchmetrics.MetricCollection,
        prog_bar: bool | None = None,
        logger: bool | None = None,
        on_step: bool | None = None,
        on_epoch: bool | None = None,
        reduce_fx: str | Callable | None = None,
        enable_graph: bool | None = None,
        sync_dist: bool | None = None,
        sync_dist_group: Any | None = None,
        add_dataloader_idx: bool | None = None,
        batch_size: int | None = None,
        rank_zero_only: bool | None = None,
    ) -> None:
        """Log a dictionary of values at once.

        Example::

            values = {'loss': loss, 'acc': acc, ..., 'metric_n': metric_n}
            self.log_dict(values)

        Args:
            dictionary: key value pairs.
                Keys must be identical across all processes if using DDP or any other distributed strategy.
                The values can be a ``float``, ``Tensor``, ``Metric``, or ``MetricCollection``.
            prog_bar: if ``True`` logs to the progress base.
            logger: if ``True`` logs to the logger.
            on_step: if ``True`` logs at this step.
                ``None`` auto-logs for training_step but not validation/test_step.
                The default value is determined by the hook.
                See :ref:`extensions/logging:Automatic Logging` for details.
            on_epoch: if ``True`` logs epoch accumulated metrics.
                ``None`` auto-logs for val/test step but not ``training_step``.
                The default value is determined by the hook.
                See :ref:`extensions/logging:Automatic Logging` for details.
            reduce_fx: reduction function over step values for end of epoch. :meth:`torch.mean` by default.
            enable_graph: if ``True``, will not auto-detach the graph
            sync_dist: if ``True``, reduces the metric across GPUs/TPUs. Use with care as this may lead to a significant
                communication overhead.
            sync_dist_group: the ddp group to sync across.
            add_dataloader_idx: if ``True``, appends the index of the current dataloader to
                the name (when using multiple). If ``False``, user needs to give unique names for
                each dataloader to not mix values.
            batch_size: Current batch size. This will be directly inferred from the loaded batch,
                but some data structures might need to explicitly provide it.
            rank_zero_only: Tells Lightning if you are calling ``self.log`` from every process (default) or only from
                rank 0. If ``True``, you won't be able to use this metric as a monitor in callbacks
                (e.g., early stopping). Warning: Improper use can lead to deadlocks! See
                :ref:`Advanced Logging <visualize/logging_advanced:rank_zero_only>` for more details.

        """

        _, fn_kwargs = self._make_prefix_and_kwargs_dict(
            _LogContextKwargs(
                prog_bar=prog_bar,
                logger=logger,
                on_step=on_step,
                on_epoch=on_epoch,
                reduce_fx=reduce_fx,
                enable_graph=enable_graph,
                sync_dist=sync_dist,
                sync_dist_group=sync_dist_group,
                add_dataloader_idx=add_dataloader_idx,
                batch_size=batch_size,
                rank_zero_only=rank_zero_only,
            )
        )
        # NOTE: Prefix will be handled by the individual log calls.
        return super().log_dict(dictionary, **fn_kwargs)



from __future__ import annotations

import logging
import os
from pathlib import Path
from typing import Literal, cast

from lightning.pytorch import LightningModule, Trainer
from lightning.pytorch.callbacks.callback import Callback
from lightning.pytorch.loggers import WandbLogger
from nshrunner._env import SNAPSHOT_DIR
from typing_extensions import final, override

from .base import CallbackConfigBase, callback_registry

log = logging.getLogger(__name__)


@final
@callback_registry.register
class WandbUploadCodeCallbackConfig(CallbackConfigBase):
    name: Literal["wandb_upload_code"] = "wandb_upload_code"

    enabled: bool = True
    """Enable uploading the code to wandb."""

    def __bool__(self):
        return self.enabled

    @override
    def create_callbacks(self, trainer_config):
        if not self:
            return

        yield WandbUploadCodeCallback(self)


class WandbUploadCodeCallback(Callback):
    def __init__(self, config: WandbUploadCodeCallbackConfig):
        super().__init__()

        self.config = config

    @override
    def setup(self, trainer: Trainer, pl_module: LightningModule, stage: str):
        if not self.config:
            return

        if not trainer.is_global_zero:
            return

        if (
            logger := next(
                (
                    logger
                    for logger in trainer.loggers
                    if isinstance(logger, WandbLogger)
                ),
                None,
            )
        ) is None:
            log.warning("Wandb logger not found. Skipping code upload.")
            return

        from wandb.wandb_run import Run

        run = cast(Run, logger.experiment)

        # If a snapshot has been taken (which can be detected using the SNAPSHOT_DIR env),
        # then upload all contents within the snapshot directory to the repository.
        if not (snapshot_dir := os.environ.get(SNAPSHOT_DIR)):
            log.debug("No snapshot directory found. Skipping upload.")
            return

        snapshot_dir = Path(snapshot_dir)
        if not snapshot_dir.exists() or not snapshot_dir.is_dir():
            log.warning(
                f"Snapshot directory '{snapshot_dir}' does not exist or is not a directory."
            )
            return

        log.info(f"Uploading code from snapshot directory '{snapshot_dir}'")
        run.log_code(str(snapshot_dir.absolute()))



from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING, Any

import torch
from lightning.pytorch import LightningModule
from lightning.pytorch.callbacks import Callback as _LightningCallback
from lightning.pytorch.utilities.types import STEP_OUTPUT
from torch.optim import Optimizer

if TYPE_CHECKING:
    from .trainer import Trainer


class NTCallbackBase(_LightningCallback):
    def setup(  # pyright: ignore[reportIncompatibleMethodOverride]
        self, trainer: Trainer, pl_module: LightningModule, stage: str
    ) -> None:
        """Called when fit, validate, test, predict, or tune begins."""

    def teardown(  # pyright: ignore[reportIncompatibleMethodOverride]
        self, trainer: Trainer, pl_module: LightningModule, stage: str
    ) -> None:
        """Called when fit, validate, test, predict, or tune ends."""

    def on_fit_start(self, trainer: Trainer, pl_module: LightningModule) -> None:  # pyright: ignore[reportIncompatibleMethodOverride]
        """Called when fit begins."""

    def on_fit_end(self, trainer: Trainer, pl_module: LightningModule) -> None:  # pyright: ignore[reportIncompatibleMethodOverride]
        """Called when fit ends."""

    def on_sanity_check_start(  # pyright: ignore[reportIncompatibleMethodOverride]
        self, trainer: Trainer, pl_module: LightningModule
    ) -> None:
        """Called when the validation sanity check starts."""

    def on_sanity_check_end(  # pyright: ignore[reportIncompatibleMethodOverride]
        self, trainer: Trainer, pl_module: LightningModule
    ) -> None:
        """Called when the validation sanity check ends."""

    def on_train_batch_start(  # pyright: ignore[reportIncompatibleMethodOverride]
        self,
        trainer: Trainer,
        pl_module: LightningModule,
        batch: Any,
        batch_idx: int,
    ) -> None:
        """Called when the train batch begins."""

    def on_train_batch_end(  # pyright: ignore[reportIncompatibleMethodOverride]
        self,
        trainer: Trainer,
        pl_module: LightningModule,
        outputs: STEP_OUTPUT,
        batch: Any,
        batch_idx: int,
    ) -> None:
        """Called when the train batch ends.

        Note:
            The value ``outputs["loss"]`` here will be the normalized value w.r.t ``accumulate_grad_batches`` of the
            loss returned from ``training_step``.

        """

    def on_train_epoch_start(  # pyright: ignore[reportIncompatibleMethodOverride]
        self, trainer: Trainer, pl_module: LightningModule
    ) -> None:
        """Called when the train epoch begins."""

    def on_train_epoch_end(  # pyright: ignore[reportIncompatibleMethodOverride]
        self, trainer: Trainer, pl_module: LightningModule
    ) -> None:
        """Called when the train epoch ends.

        To access all batch outputs at the end of the epoch, you can cache step outputs as an attribute of the
        :class:`lightning.pytorch.core.LightningModule` and access them in this hook:

        .. code-block:: python

            class MyLightningModule(L.LightningModule):
                 def __init__(self):
                    super().__init__() # pyright: ignore[reportIncompatibleMethodOverride]
                    self.training_step_outputs = []

                 def training_step(self):
                    loss = ... # pyright: ignore[reportIncompatibleMethodOverride]
                    self.training_step_outputs.append(loss)
                    return loss


            class MyCallback(L.Callback):
                 def on_train_epoch_end(self, trainer, pl_module):
                    # do something with all training_step outputs, for example: # pyright: ignore[reportIncompatibleMethodOverride]
                    epoch_mean = torch.stack(pl_module.training_step_outputs).mean()
                    pl_module.log("training_epoch_mean", epoch_mean)
                    # free up the memory
                    pl_module.training_step_outputs.clear()

        """

    def on_validation_epoch_start(  # pyright: ignore[reportIncompatibleMethodOverride]
        self, trainer: Trainer, pl_module: LightningModule
    ) -> None:
        """Called when the val epoch begins."""

    def on_validation_epoch_end(  # pyright: ignore[reportIncompatibleMethodOverride]
        self, trainer: Trainer, pl_module: LightningModule
    ) -> None:
        """Called when the val epoch ends."""

    def on_test_epoch_start(  # pyright: ignore[reportIncompatibleMethodOverride]
        self, trainer: Trainer, pl_module: LightningModule
    ) -> None:
        """Called when the test epoch begins."""

    def on_test_epoch_end(  # pyright: ignore[reportIncompatibleMethodOverride]
        self, trainer: Trainer, pl_module: LightningModule
    ) -> None:
        """Called when the test epoch ends."""

    def on_predict_epoch_start(  # pyright: ignore[reportIncompatibleMethodOverride]
        self, trainer: Trainer, pl_module: LightningModule
    ) -> None:
        """Called when the predict epoch begins."""

    def on_predict_epoch_end(  # pyright: ignore[reportIncompatibleMethodOverride]
        self, trainer: Trainer, pl_module: LightningModule
    ) -> None:
        """Called when the predict epoch ends."""

    def on_validation_batch_start(  # pyright: ignore[reportIncompatibleMethodOverride]
        self,
        trainer: Trainer,
        pl_module: LightningModule,
        batch: Any,
        batch_idx: int,
        dataloader_idx: int = 0,
    ) -> None:
        """Called when the validation batch begins."""

    def on_validation_batch_end(  # pyright: ignore[reportIncompatibleMethodOverride]
        self,
        trainer: Trainer,
        pl_module: LightningModule,
        outputs: STEP_OUTPUT,
        batch: Any,
        batch_idx: int,
        dataloader_idx: int = 0,
    ) -> None:
        """Called when the validation batch ends."""

    def on_test_batch_start(  # pyright: ignore[reportIncompatibleMethodOverride]
        self,
        trainer: Trainer,
        pl_module: LightningModule,
        batch: Any,
        batch_idx: int,
        dataloader_idx: int = 0,
    ) -> None:
        """Called when the test batch begins."""

    def on_test_batch_end(  # pyright: ignore[reportIncompatibleMethodOverride]
        self,
        trainer: Trainer,
        pl_module: LightningModule,
        outputs: STEP_OUTPUT,
        batch: Any,
        batch_idx: int,
        dataloader_idx: int = 0,
    ) -> None:
        """Called when the test batch ends."""

    def on_predict_batch_start(  # pyright: ignore[reportIncompatibleMethodOverride]
        self,
        trainer: Trainer,
        pl_module: LightningModule,
        batch: Any,
        batch_idx: int,
        dataloader_idx: int = 0,
    ) -> None:
        """Called when the predict batch begins."""

    def on_predict_batch_end(  # pyright: ignore[reportIncompatibleMethodOverride]
        self,
        trainer: Trainer,
        pl_module: LightningModule,
        outputs: Any,
        batch: Any,
        batch_idx: int,
        dataloader_idx: int = 0,
    ) -> None:
        """Called when the predict batch ends."""

    def on_train_start(self, trainer: Trainer, pl_module: LightningModule) -> None:  # pyright: ignore[reportIncompatibleMethodOverride]
        """Called when the train begins."""

    def on_train_end(self, trainer: Trainer, pl_module: LightningModule) -> None:  # pyright: ignore[reportIncompatibleMethodOverride]
        """Called when the train ends."""

    def on_validation_start(  # pyright: ignore[reportIncompatibleMethodOverride]
        self, trainer: Trainer, pl_module: LightningModule
    ) -> None:
        """Called when the validation loop begins."""

    def on_validation_end(  # pyright: ignore[reportIncompatibleMethodOverride]
        self, trainer: Trainer, pl_module: LightningModule
    ) -> None:
        """Called when the validation loop ends."""

    def on_test_start(self, trainer: Trainer, pl_module: LightningModule) -> None:  # pyright: ignore[reportIncompatibleMethodOverride]
        """Called when the test begins."""

    def on_test_end(self, trainer: Trainer, pl_module: LightningModule) -> None:  # pyright: ignore[reportIncompatibleMethodOverride]
        """Called when the test ends."""

    def on_predict_start(  # pyright: ignore[reportIncompatibleMethodOverride]
        self, trainer: Trainer, pl_module: LightningModule
    ) -> None:
        """Called when the predict begins."""

    def on_predict_end(self, trainer: Trainer, pl_module: LightningModule) -> None:  # pyright: ignore[reportIncompatibleMethodOverride]
        """Called when predict ends."""

    def on_exception(  # pyright: ignore[reportIncompatibleMethodOverride]
        self,
        trainer: Trainer,
        pl_module: LightningModule,
        exception: BaseException,
    ) -> None:
        """Called when any trainer execution is interrupted by an exception."""

    def state_dict(self) -> dict[str, Any]:  # pyright: ignore[reportIncompatibleMethodOverride]
        """Called when saving a checkpoint, implement to generate callback's ``state_dict``.

        Returns:
            A dictionary containing callback state.

        """
        return {}

    def load_state_dict(self, state_dict: dict[str, Any]) -> None:  # pyright: ignore[reportIncompatibleMethodOverride]
        """Called when loading a checkpoint, implement to reload callback state given callback's ``state_dict``.

        Args:
            state_dict: the callback state returned by ``state_dict``.

        """
        pass

    def on_save_checkpoint(  # pyright: ignore[reportIncompatibleMethodOverride]
        self,
        trainer: Trainer,
        pl_module: LightningModule,
        checkpoint: dict[str, Any],
    ) -> None:
        r"""Called when saving a checkpoint to give you a chance to store anything else you might want to save.

        Args:
            trainer: the current :class:`~lightning.pytorch.trainer.trainer.Trainer` instance.
            pl_module: the current :class:`~lightning.pytorch.core.LightningModule` instance.
            checkpoint: the checkpoint dictionary that will be saved.

        """

    def on_load_checkpoint(  # pyright: ignore[reportIncompatibleMethodOverride]
        self,
        trainer: Trainer,
        pl_module: LightningModule,
        checkpoint: dict[str, Any],
    ) -> None:
        r"""Called when loading a model checkpoint, use to reload state.

        Args:
            trainer: the current :class:`~lightning.pytorch.trainer.trainer.Trainer` instance.
            pl_module: the current :class:`~lightning.pytorch.core.LightningModule` instance.
            checkpoint: the full checkpoint dictionary that got loaded by the Trainer.

        """

    def on_before_backward(  # pyright: ignore[reportIncompatibleMethodOverride]
        self, trainer: Trainer, pl_module: LightningModule, loss: torch.Tensor
    ) -> None:
        """Called before ``loss.backward()``."""

    def on_after_backward(  # pyright: ignore[reportIncompatibleMethodOverride]
        self, trainer: Trainer, pl_module: LightningModule
    ) -> None:
        """Called after ``loss.backward()`` and before optimizers are stepped."""

    def on_before_optimizer_step(  # pyright: ignore[reportIncompatibleMethodOverride]
        self,
        trainer: Trainer,
        pl_module: LightningModule,
        optimizer: Optimizer,
    ) -> None:
        """Called before ``optimizer.step()``."""

    def on_before_zero_grad(  # pyright: ignore[reportIncompatibleMethodOverride]
        self,
        trainer: Trainer,
        pl_module: LightningModule,
        optimizer: Optimizer,
    ) -> None:
        """Called before ``optimizer.zero_grad()``."""

    def on_checkpoint_saved(  # pyright: ignore[reportIncompatibleMethodOverride]
        self,
        ckpt_path: Path,
        metadata_path: Path | None,
        trainer: Trainer,
        pl_module: LightningModule,
    ) -> None:
        """Called after a checkpoint is saved."""
        pass


def _call_on_checkpoint_saved(
    trainer: Trainer,
    ckpt_path: str | Path,
    metadata_path: str | Path | None,
):
    ckpt_path = Path(ckpt_path)
    metadata_path = Path(metadata_path) if metadata_path else None

    for callback in trainer.callbacks:
        if not isinstance(callback, NTCallbackBase):
            continue

        callback.on_checkpoint_saved(
            ckpt_path,
            metadata_path,
            trainer,
            trainer.lightning_module,
        )



from __future__ import annotations

import contextlib

import nshconfig as C
import torch


@contextlib.contextmanager
def rng_context(config: RNGConfig | None):
    with contextlib.ExitStack() as stack:
        if config is not None:
            stack.enter_context(
                torch.random.fork_rng(devices=range(torch.cuda.device_count()))
            )
            torch.manual_seed(config.seed)

        yield


class RNGConfig(C.Config):
    seed: int
    """Random seed to use for initialization."""



from __future__ import annotations

import copy
import datetime
import logging
from pathlib import Path
from typing import TYPE_CHECKING, Any, ClassVar

import nshconfig as C
import numpy as np
import torch

from ..util._environment_info import EnvironmentConfig
from ..util.path import compute_file_checksum, path_exists, try_symlink_or_copy

if TYPE_CHECKING:
    from ..trainer.trainer import Trainer

try:
    from pydantic import BaseModel

    _HAS_PYDANTIC = True
except ImportError:
    if not TYPE_CHECKING:
        BaseModel = object
    else:
        from pydantic import BaseModel
    _HAS_PYDANTIC = False


log = logging.getLogger(__name__)


METADATA_PATH_SUFFIX = ".metadata.json"


def _full_hparams_dict(trainer: Trainer):
    hparams = {}
    hparams["trainer"] = trainer.hparams.model_dump(mode="json")

    if trainer.lightning_module is not None:
        model_hparams = trainer.lightning_module.hparams
        if _HAS_PYDANTIC and isinstance(model_hparams, BaseModel):
            model_hparams = model_hparams.model_dump(mode="json")
        hparams["model"] = dict(model_hparams)

    return hparams


class CheckpointMetadata(C.Config):
    PATH_SUFFIX: ClassVar[str] = METADATA_PATH_SUFFIX

    checkpoint_path: Path
    checkpoint_filename: str
    checkpoint_checksum: str

    run_id: str
    name: str
    project: str | None
    checkpoint_timestamp: datetime.datetime
    start_timestamp: datetime.datetime | None

    epoch: int
    global_step: int
    training_time: datetime.timedelta
    metrics: dict[str, Any]
    environment: EnvironmentConfig

    hparams: Any | None

    @classmethod
    def from_file(cls, path: Path):
        return cls.model_validate_json(path.read_text(encoding="utf-8"))

    @classmethod
    def from_ckpt_path(cls, checkpoint_path: Path):
        if not (metadata_path := checkpoint_path.with_suffix(cls.PATH_SUFFIX)).exists():
            raise FileNotFoundError(
                f"Metadata file not found for checkpoint: {checkpoint_path}"
            )
        return cls.from_file(metadata_path)


def _generate_checkpoint_metadata(
    trainer: Trainer,
    checkpoint_path: Path,
    metadata_path: Path,
):
    checkpoint_timestamp = datetime.datetime.now()
    start_timestamp = trainer.start_time()
    training_time = trainer.time_elapsed()

    metrics: dict[str, Any] = {}
    for name, metric in copy.deepcopy(trainer.callback_metrics).items():
        match metric:
            case torch.Tensor() | np.ndarray():
                metrics[name] = metric.detach().cpu().item()
            case _:
                metrics[name] = metric

    return CheckpointMetadata(
        # checkpoint_path=checkpoint_path,
        # We should store the path as a relative path
        # to the metadata file to avoid issues with
        # moving the checkpoint directory
        checkpoint_path=checkpoint_path.relative_to(metadata_path.parent),
        checkpoint_filename=checkpoint_path.name,
        checkpoint_checksum=compute_file_checksum(checkpoint_path),
        run_id=trainer.hparams.id,
        name=trainer.hparams.full_name,
        project=trainer.hparams.project,
        checkpoint_timestamp=checkpoint_timestamp,
        start_timestamp=start_timestamp.datetime
        if start_timestamp is not None
        else None,
        epoch=trainer.current_epoch,
        global_step=trainer.global_step,
        training_time=training_time,
        metrics=metrics,
        environment=trainer.hparams.environment,
        hparams=_full_hparams_dict(trainer),
    )


def _metadata_path(checkpoint_path: Path):
    return checkpoint_path.with_suffix(CheckpointMetadata.PATH_SUFFIX)


def write_checkpoint_metadata(trainer: Trainer, checkpoint_path: Path):
    metadata_path = _metadata_path(checkpoint_path)
    metadata = _generate_checkpoint_metadata(trainer, checkpoint_path, metadata_path)

    # Write the metadata to the checkpoint directory
    try:
        metadata_path.write_text(metadata.model_dump_json(indent=4), encoding="utf-8")
    except Exception:
        log.warning(f"Failed to write metadata to {metadata_path}", exc_info=True)
        return None

    log.debug(f"Checkpoint metadata written to {metadata_path}")
    return metadata_path


def remove_checkpoint_metadata(checkpoint_path: Path):
    path = _metadata_path(checkpoint_path)
    try:
        path.unlink(missing_ok=True)
    except Exception:
        log.warning(f"Failed to remove {path}", exc_info=True)
    else:
        log.debug(f"Removed {path}")


def remove_checkpoint_metadata_link(ckpt_link_path: Path):
    path = _metadata_path(ckpt_link_path)
    # If the metadata does not exist, we can safely ignore this
    if not path_exists(path, follow_symlinks=False):
        # This is EXTREMELY important here
        # Otherwise, we've already deleted the file that the symlink
        # used to point to, so this always returns False
        log.debug(f"Metadata file does not exist: {path}")
        return

    # If the metadata exists, we can remove it
    try:
        path.unlink(missing_ok=True)
    except Exception:
        log.warning(f"Failed to remove {path}", exc_info=True)
    else:
        log.debug(f"Removed {path}")


def link_checkpoint_metadata(checkpoint_path: Path, linked_checkpoint_path: Path):
    # First, remove any existing metadata files
    remove_checkpoint_metadata_link(linked_checkpoint_path)

    # Link the metadata files to the new checkpoint
    path = _metadata_path(checkpoint_path)
    linked_path = _metadata_path(linked_checkpoint_path)
    try_symlink_or_copy(path, linked_path)



from __future__ import annotations

import logging
from typing import Literal

from typing_extensions import final, override

from .._callback import NTCallbackBase
from .base import CallbackConfigBase, callback_registry

log = logging.getLogger(__name__)


@final
@callback_registry.register
class DebugFlagCallbackConfig(CallbackConfigBase):
    name: Literal["debug_flag"] = "debug_flag"

    enabled: bool = True
    """Whether to enable the callback."""

    def __bool__(self):
        return self.enabled

    @override
    def create_callbacks(self, trainer_config):
        if not self:
            return

        yield DebugFlagCallback(self)


class DebugFlagCallback(NTCallbackBase):
    """
    Sets the debug flag to true in the following circumstances:
    - fast_dev_run is enabled
    - sanity check is running
    """

    @override
    def __init__(self, config: DebugFlagCallbackConfig):
        super().__init__()

        self.config = config
        del config

        self._debug = False

    @override
    def setup(self, trainer, pl_module, stage):
        if not getattr(trainer, "fast_dev_run", False):
            return

        if not trainer.debug:
            log.critical("Fast dev run detected, setting debug flag to True.")
        trainer.debug = True

    @override
    def on_sanity_check_start(self, trainer, pl_module):
        self._debug = trainer.debug
        if not self._debug:
            log.critical("Enabling debug flag during sanity check routine.")
        trainer.debug = True

    @override
    def on_sanity_check_end(self, trainer, pl_module):
        if not self._debug:
            log.critical("Sanity check routine complete, disabling debug flag.")
        trainer.debug = self._debug



from __future__ import annotations

from typing import Literal

from typing_extensions import override

from .base import PluginConfigBase, plugin_registry


@plugin_registry.register
class TorchCheckpointIOPlugin(PluginConfigBase):
    """CheckpointIO that utilizes torch.save and torch.load to save and load checkpoints respectively."""

    name: Literal["torch_checkpoint"] = "torch_checkpoint"

    @override
    def create_plugin(self, trainer_config):
        from lightning.fabric.plugins.io.torch_io import TorchCheckpointIO

        return TorchCheckpointIO()


@plugin_registry.register
class XLACheckpointIOPlugin(PluginConfigBase):
    """CheckpointIO that utilizes xm.save to save checkpoints for TPU training strategies."""

    name: Literal["xla_checkpoint"] = "xla_checkpoint"

    @override
    def create_plugin(self, trainer_config):
        from lightning.fabric.plugins.io.xla import XLACheckpointIO

        return XLACheckpointIO()


@plugin_registry.register
class AsyncCheckpointIOPlugin(PluginConfigBase):
    """Enables saving the checkpoints asynchronously in a thread.

    .. warning::  This is an experimental feature.
    """

    name: Literal["async_checkpoint"] = "async_checkpoint"

    checkpoint_io: TorchCheckpointIOPlugin | None = None
    """A checkpoint IO plugin that is used as the basis for async checkpointing."""

    @override
    def create_plugin(self, trainer_config):
        from lightning.pytorch.plugins.io import AsyncCheckpointIO, CheckpointIO

        base_io = (
            self.checkpoint_io.create_plugin(trainer_config)
            if self.checkpoint_io
            else None
        )
        if base_io is not None and not isinstance(base_io, CheckpointIO):
            raise TypeError(
                f"Expected `checkpoint_io` to be a `CheckpointIO` instance, but got {type(base_io)}."
            )
        return AsyncCheckpointIO(checkpoint_io=base_io)



from __future__ import annotations

from typing import Literal

from typing_extensions import final, override

from .base import LoggerConfigBase, logger_registry


@final
@logger_registry.register
class CSVLoggerConfig(LoggerConfigBase):
    name: Literal["csv"] = "csv"

    enabled: bool = True
    """Enable CSV logging."""

    priority: int = 0
    """Priority of the logger. Higher priority loggers are created first."""

    prefix: str = ""
    """A string to put at the beginning of metric keys."""

    flush_logs_every_n_steps: int = 100
    """How often to flush logs to disk."""

    @override
    def create_logger(self, trainer_config):
        if not self.enabled:
            return None

        from lightning.pytorch.loggers.csv_logs import CSVLogger

        save_dir = trainer_config.directory._resolve_log_directory_for_logger(
            trainer_config.id,
            self,
        )
        return CSVLogger(
            save_dir=save_dir,
            name=trainer_config.full_name,
            version=trainer_config.id,
            prefix=self.prefix,
            flush_logs_every_n_steps=self.flush_logs_every_n_steps,
        )



from __future__ import annotations

from argparse import Namespace
from typing import Any, Literal

import numpy as np
from lightning.pytorch.loggers import Logger
from typing_extensions import final

from .base import LoggerConfigBase, logger_registry


@final
@logger_registry.register
class ActSaveLoggerConfig(LoggerConfigBase):
    name: Literal["actsave"] = "actsave"

    def create_logger(self, trainer_config):
        if not self.enabled:
            return None

        return ActSaveLogger()


class ActSaveLogger(Logger):
    @property
    def name(self):
        return None

    @property
    def version(self):
        from nshutils import ActSave

        if ActSave._saver is None:
            return None

        return ActSave._saver._id

    @property
    def save_dir(self):
        from nshutils import ActSave

        if ActSave._saver is None:
            return None

        return str(ActSave._saver._save_dir)

    def log_hyperparams(
        self,
        params: dict[str, Any] | Namespace,
        *args: Any,
        **kwargs: Any,
    ):
        from nshutils import ActSave

        # Wrap the hparams as a object-dtype np array
        return ActSave.save({"hyperparameters": np.array(params, dtype=object)})

    def log_metrics(self, metrics: dict[str, float], step: int | None = None) -> None:
        from nshutils import ActSave

        ActSave.save({**metrics})



from __future__ import annotations

import contextlib
import copy
from collections.abc import Callable, Sequence
from typing import Any, Literal, Protocol, runtime_checkable

import nshconfig as C
import torch
import torch.nn as nn
from typing_extensions import deprecated, override

from .nonlinearity import NonlinearityConfig, NonlinearityConfigBase


@runtime_checkable
class LinearModuleConstructor(Protocol):
    def __call__(
        self, in_features: int, out_features: int, bias: bool = True
    ) -> nn.Module: ...


class ResidualSequential(nn.Sequential):
    @override
    def forward(self, input: torch.Tensor) -> torch.Tensor:
        return input + super().forward(input)


class MLPConfig(C.Config):
    bias: bool = True
    """Whether to include bias terms in the linear layers."""

    no_bias_scalar: bool = True
    """Whether to exclude bias terms when the output dimension is 1."""

    nonlinearity: NonlinearityConfig | None = None
    """Activation function to use between layers."""

    ln: bool | Literal["pre", "post"] = False
    """Whether to apply layer normalization before or after the linear layers."""

    dropout: float | None = None
    """Dropout probability to apply between layers."""

    residual: bool = False
    """Whether to use residual connections between layers."""

    seed: int | None = None
    """Random seed to use for initialization. If None, the default Torch behavior is used."""

    @deprecated("Use `nt.nn.MLP(config=...)` instead.")
    def create_module(
        self,
        dims: Sequence[int],
        pre_layers: Sequence[nn.Module] = [],
        post_layers: Sequence[nn.Module] = [],
        linear_cls: LinearModuleConstructor = nn.Linear,
    ):
        kwargs: dict[str, Any] = {
            "bias": self.bias,
            "no_bias_scalar": self.no_bias_scalar,
            "nonlinearity": self.nonlinearity,
            "ln": self.ln,
            "dropout": self.dropout,
            "residual": self.residual,
            "seed": self.seed,
        }
        return MLP(
            dims,
            **kwargs,
            pre_layers=pre_layers,
            post_layers=post_layers,
            linear_cls=linear_cls,
        )


@contextlib.contextmanager
def custom_seed_context(seed: int | None):
    with contextlib.ExitStack() as stack:
        if seed is not None:
            stack.enter_context(
                torch.random.fork_rng(devices=range(torch.cuda.device_count()))
            )
            torch.manual_seed(seed)

        yield


def MLP(
    dims: Sequence[int],
    activation: NonlinearityConfigBase
    | nn.Module
    | Callable[[], nn.Module]
    | None = None,
    nonlinearity: NonlinearityConfigBase
    | nn.Module
    | Callable[[], nn.Module]
    | None = None,
    bias: bool | None = None,
    no_bias_scalar: bool | None = None,
    ln: bool | Literal["pre", "post"] | None = None,
    dropout: float | None = None,
    residual: bool | None = None,
    pre_layers: Sequence[nn.Module] = [],
    post_layers: Sequence[nn.Module] = [],
    linear_cls: LinearModuleConstructor = nn.Linear,
    seed: int | None = None,
    config: MLPConfig | None = None,
):
    """
    Constructs a multi-layer perceptron (MLP) with the given dimensions and activation function.

    Args:
        dims (Sequence[int]): List of integers representing the dimensions of the MLP.
        nonlinearity (Callable[[], nn.Module] | None, optional): Activation function to use between layers.
        activation (Callable[[], nn.Module] | None, optional): Activation function to use between layers.
        bias (bool | None, optional): Whether to include bias terms in the linear layers.
        no_bias_scalar (bool | None, optional): Whether to exclude bias terms when the output dimension is 1.
        ln (bool | Literal["pre", "post"] | None, optional): Whether to apply layer normalization before or after the linear layers.
        dropout (float | None, optional): Dropout probability to apply between layers.
        residual (bool | None, optional): Whether to use residual connections between layers.
        pre_layers (Sequence[nn.Module], optional): List of layers to insert before the linear layers. Defaults to [].
        post_layers (Sequence[nn.Module], optional): List of layers to insert after the linear layers. Defaults to [].
        linear_cls (LinearModuleConstructor, optional): Linear module constructor to use. Defaults to nn.Linear.
        seed (int | None, optional): Random seed to use for initialization. If None, the default Torch behavior is used.
        config (MLPConfig | None, optional): Configuration object for the MLP. Parameters specified directly take precedence.

    Returns:
        nn.Sequential: The constructed MLP.
    """

    # Resolve parameters: arg if not None, otherwise config value if config exists, otherwise default
    resolved_bias = bias if bias is not None else (config.bias if config else True)
    resolved_no_bias_scalar = (
        no_bias_scalar
        if no_bias_scalar is not None
        else (config.no_bias_scalar if config else True)
    )
    resolved_nonlinearity = (
        nonlinearity
        if nonlinearity is not None
        else (config.nonlinearity if config else None)
    )
    resolved_ln = ln if ln is not None else (config.ln if config else False)
    resolved_dropout = (
        dropout if dropout is not None else (config.dropout if config else None)
    )
    resolved_residual = (
        residual if residual is not None else (config.residual if config else False)
    )
    resolved_seed = seed if seed is not None else (config.seed if config else None)

    with custom_seed_context(resolved_seed):
        if activation is None:
            activation = resolved_nonlinearity

        if len(dims) < 2:
            raise ValueError("mlp requires at least 2 dimensions")
        if resolved_ln is True:
            resolved_ln = "pre"
        elif isinstance(resolved_ln, str) and resolved_ln not in ("pre", "post"):
            raise ValueError("ln must be a boolean or 'pre' or 'post'")

        layers: list[nn.Module] = []
        if resolved_ln == "pre":
            layers.append(nn.LayerNorm(dims[0]))

        layers.extend(pre_layers)

        for i in range(len(dims) - 1):
            in_features = dims[i]
            out_features = dims[i + 1]
            bias_ = resolved_bias and not (
                resolved_no_bias_scalar and out_features == 1
            )
            layers.append(linear_cls(in_features, out_features, bias=bias_))
            if resolved_dropout is not None:
                layers.append(nn.Dropout(resolved_dropout))
            if i < len(dims) - 2:
                match activation:
                    case NonlinearityConfigBase():
                        layers.append(activation.create_module())
                    case nn.Module():
                        # In this case, we create a deep copy of the module to avoid sharing parameters (if any).
                        layers.append(copy.deepcopy(activation))
                    case Callable():
                        layers.append(activation())
                    case _:
                        raise ValueError(
                            "Either `nonlinearity` or `activation` must be provided"
                        )

        layers.extend(post_layers)

        if resolved_ln == "post":
            layers.append(nn.LayerNorm(dims[-1]))

        cls = ResidualSequential if resolved_residual else nn.Sequential
        return cls(*layers)
